
#include <Y/y.h>
#include <Y/object/parameters.h>
#include <Y/util/yutil.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

struct ParametersEntry
{
  const char *c;
  struct ParametersEntry *next;
};

struct ParametersFree
{
  char *c;
  struct ParametersFree *next;
};

struct Parameters
{
  int inputCount, outputCount, error;
  int inputShift;
  struct ParametersEntry *firstInput, *firstOutput;
  struct ParametersEntry *lastInput, *lastOutput;
  struct ParametersFree *firstFree;
  const char **inputStrings;
  const char **outputStrings;
};

struct Parameters *
parametersCreate ()
{
  struct Parameters *params = ymalloc (sizeof (struct Parameters));
  params -> inputCount = 0;
  params -> outputCount = 0;
  params -> error = 0;
  params -> inputShift = 0;
  params -> firstInput = NULL;
  params -> lastInput = NULL;
  params -> firstOutput = NULL;
  params -> lastOutput = NULL;
  params -> firstFree = NULL;
  params -> inputStrings = NULL;
  params -> outputStrings = NULL;
  return params;
}

struct Parameters *
parametersBuild (char *data, int len)
{
  /* builds parameters out of \0-separated strings */
  /* turn \x1c-seperated strings into \0-seperated strings */
  struct Parameters *params = parametersCreate ();
  char *a = data;
  char *p = data;
  int c = 0;
  while (c < len)
    {
      if (*p == '\x1c' || *p == '\0')
        {
          *p = '\0';
          parametersAddInput (params, a);
          a = p + 1;
        }
      ++p; ++c;
    }
  return params;
}

struct Parameters *
parametersDuplicate (struct Parameters *orig)
{
  struct Parameters *self = parametersCreate ();
  struct ParametersEntry *currentOrig, **currentSelf_p;
  self -> inputCount = orig -> inputCount;
  self -> outputCount = orig -> outputCount;
  self -> error = orig -> error;

  currentOrig = orig -> firstInput;
  currentSelf_p = &(self -> firstInput);
  while (currentOrig != NULL)
    {
      (*currentSelf_p) = ymalloc (sizeof (struct ParametersEntry));
      (*currentSelf_p) -> c = currentOrig -> c;
      (*currentSelf_p) -> next = NULL;
      currentSelf_p = &((*currentSelf_p) -> next);
      currentOrig = currentOrig -> next;
    }

  currentOrig = orig -> firstOutput;
  currentSelf_p = &(self -> firstOutput);
  while (currentOrig != NULL)
    {
      (*currentSelf_p) = ymalloc (sizeof (struct ParametersEntry));
      (*currentSelf_p) -> c = currentOrig -> c;
      (*currentSelf_p) -> next = NULL;
      currentSelf_p = &((*currentSelf_p) -> next);
      currentOrig = currentOrig -> next;
    }

   return self; 
}

static void
parametersFreeze (struct Parameters *self, int count,
                  struct ParametersEntry **first, const char *** refs)
{
  int i = 0;
  struct ParametersEntry * current = *first;
  if (*refs != NULL)
    return;
  *refs = ymalloc (sizeof (const char *) * count);
  while (current != NULL)
    {
      (*refs)[i] = current -> c;
      current = current -> next;
      ++i;
    }
  return;
}

static void
parametersThaw (struct Parameters *self, const char *** refs)
{
  if (*refs == NULL)
    return;
  yfree (*refs);
  *refs = NULL;
}

void
parametersAddInput (struct Parameters *self, const char *string)
{
  struct ParametersEntry *newEntry = ymalloc (sizeof (struct ParametersEntry));
  newEntry -> c = string;
  newEntry -> next = NULL;

  parametersThaw (self, &(self -> inputStrings));

  if (self -> lastInput == NULL)
    self -> firstInput = newEntry;
  else
    self -> lastInput -> next = newEntry;
  self -> lastInput = newEntry;

  self -> inputCount ++;
}

void
parametersAddInputInteger (struct Parameters *self, int value)
{
  /* a 128 bit number can be no longer than 40 decimal digits...
   * we use 64 to be sure.
   */

  char *buffer = ymalloc (64);
  snprintf (buffer, 64, "%d", value);
  buffer[63] = '\0';
  parametersAddInput (self, buffer);
  parametersFreeOnDestroy (self, buffer); 
}

int
parametersGetInputCount (struct Parameters *self)
{
  return self -> inputCount - self -> inputShift;
}

const char *
parametersGetInput (struct Parameters *self, int n)
{
  n += self -> inputShift;
  if (n >= self -> inputCount)
    return NULL;
  parametersFreeze (self, self -> inputCount, &(self -> firstInput),
                    &(self -> inputStrings));
  return self -> inputStrings[n];
}

void
parametersShiftInput (struct Parameters *self, int amount)
{
  self -> inputShift += amount;
  if (self -> inputShift < 0)
    self -> inputShift = 0;
}

void
parametersAddOutput (struct Parameters *self, const char *string)
{
  struct ParametersEntry *newEntry = ymalloc (sizeof (struct ParametersEntry));
  newEntry -> c = string;
  newEntry -> next = NULL;

  parametersThaw (self, &(self -> outputStrings));

  if (self -> lastOutput == NULL)
    self -> firstOutput = newEntry;
  else
    self -> lastOutput -> next = newEntry;
  self -> lastOutput = newEntry;

  self -> outputCount ++;
}

void
parametersAddOutputInteger (struct Parameters *self, int value)
{
  /* a 128 bit number can be no longer than 40 decimal digits...
   * we use 64 to be sure.
   */

  char *buffer = ymalloc (64);
  snprintf (buffer, 64, "%d", value);
  buffer[63] = '\0';
  parametersAddOutput (self, buffer);
  parametersFreeOnDestroy (self, buffer); 
}

int
parametersGetOutputCount (struct Parameters *self)
{
  return self -> outputCount;
}

const char *
parametersGetOutput (struct Parameters *self, int n)
{
  if (n >= self -> outputCount)
    return NULL;
  parametersFreeze (self, self -> outputCount, &(self -> firstOutput),
                    &(self -> outputStrings));
  return self -> outputStrings[n];
}

char *
parametersGetOutputData (struct Parameters *self, const char *prefix,
                         int *len_p)
{
  struct ParametersEntry * current = self -> firstOutput;
  int len = 0, i = 0, prefixLen = 0;
  char *data, *cp;
  int lengths[self -> outputCount];
  if (prefix != NULL)
    {
      prefixLen = strlen (prefix) + 1;
    }
  if (prefix == NULL && self -> outputCount == 0)
    {
      if (len_p != NULL)
        *len_p = 0;
      return NULL;
    }

  while (current != NULL)
    {
      lengths[i] = strlen (current -> c) ;
      len += lengths[i] + 1;
      i ++;
      current = current -> next;
    }

  len += prefixLen;

  data = ymalloc (sizeof (char) * len);
  cp = data;

  if (prefix != NULL)
    {
      strcpy (cp, prefix);
      cp += prefixLen - 1;
      *(cp++) = '\x1c';
    }

  i = 0;
  current = self -> firstOutput;
  while (current != NULL)
    {
      strcpy (cp, current -> c);
      cp += lengths[i];
      *(cp++) = '\x1c';
      i++;
      current = current -> next;
    }
  *(--cp) = '\0';
  if (len_p != NULL)
    *len_p = len;
  return data;
  
}

void
parametersFreeOnDestroy (struct Parameters *self, char *c)
{
  struct ParametersFree *f = ymalloc (sizeof (struct ParametersFree));
  f -> c = c;
  f -> next = self -> firstFree;
  self -> firstFree = f;
}

void
parametersSetError (struct Parameters *self)
{
  self -> error = 1;
}

void
parametersClearError (struct Parameters *self)
{
  self -> error = 0;
}

int
parametersIsError (struct Parameters *self)
{
  return self -> error;
}

void
parametersDestroy (struct Parameters *self)
{
  struct ParametersEntry *currentEntry;
  struct ParametersFree *currentFree;
  parametersThaw (self, &(self -> outputStrings));
  parametersThaw (self, &(self -> inputStrings));
  currentEntry = self -> firstInput;
  while (currentEntry != NULL)
    {
      struct ParametersEntry *nextEntry = currentEntry -> next;
      yfree (currentEntry);
      currentEntry = nextEntry;
    }
  currentEntry = self -> firstOutput;
  while (currentEntry != NULL)
    {
      struct ParametersEntry *nextEntry = currentEntry -> next;
      yfree (currentEntry);
      currentEntry = nextEntry;
    }
  currentFree = self -> firstFree;
  while (currentFree != NULL)
    {
      struct ParametersFree *nextFree = currentFree -> next;
      yfree (currentFree -> c);
      yfree (currentFree);
      currentFree = nextFree;
    }
  yfree (self);
}

/* arch-tag: 83a88bd4-a59d-42be-b4cb-7f4fa41a6eab
 */
