#include <Y/modules/module_interface.h>

#include <Y/message/message.h>
#include <Y/message/client.h>
#include <Y/message/client_p.h>

#include <Y/util/yutil.h>
#include <Y/util/list.h>

#include <Y/main/control.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <netinet/in.h>

#define MESSAGE_HEADER_SIZE (4 * 8)

struct UnixData
{
  char *path;
  int listeningFD;
};

struct UnixClient
{
  struct Client client;
  int fd;
  struct Message *sending;
  struct Message *receiving;
  ssize_t sent, received;
  struct List *sendQueue;
};

static void unixSendMessage (struct Client *c, struct Message *m);
static void unixClose (struct Client *c);

struct ClientClass unixClientClass =
{
  name: "Unix Domain Socket Client",
  sendMessage: unixSendMessage,
  close: unixClose
};

static struct UnixClient *
castBack (struct Client *self_c)
{
  /* assert ( self_c -> c == &unixClientClass ); */
  return (struct UnixClient *)self_c;
}

static void
unixSendMessage (struct Client *self_c, struct Message *m)
{
  struct UnixClient *self = castBack (self_c);
  if (self -> sending == NULL)
    {
      self -> sending = m;
      self -> sent = 0;
      controlChangeFileDescriptorMask (self -> fd,
                 CONTROL_WATCH_WRITE | CONTROL_WATCH_READ );
    }
  else
    {
      listAppend (self -> sendQueue, m);
    }
}

static void
unixClose (struct Client *self_c)
{
  struct UnixClient *self = castBack (self_c);
  controlUnregisterFileDescriptor (self -> fd);
  close (self -> fd);
  listDestroy (self -> sendQueue, (void (*)(void *))messageDestroy);
  messageDestroy (self -> sending);
  messageDestroy (self -> receiving);
  yfree (self);
}



static void
unixClientReady (int fd, int causeMask, void *data_v)
{
  struct UnixClient *self = data_v;
  if (causeMask & CONTROL_WATCH_READ)
    {
      if (self -> receiving == NULL)
        {
          self -> receiving = ymalloc (sizeof (struct Message));
          self -> receiving -> data = ymalloc (MESSAGE_HEADER_SIZE);
          self -> received = 0;
        }
      if (self -> received < MESSAGE_HEADER_SIZE)
        {
          /* attempt to read the rest of the header */
          ssize_t ret = read (self -> fd,
                              self -> receiving -> data + self -> received,
                              MESSAGE_HEADER_SIZE - self -> received);
          if (ret < 1)
            {
              clientClose (&(self -> client));
              return;
            }
          self -> received += ret;
          if (self -> received == MESSAGE_HEADER_SIZE)
            {
              /* header fully received, translate */
              uint32_t *netdata = (uint32_t *)self -> receiving -> data;

              self -> receiving -> to   = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> from = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> cid  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> oid  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> op   = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> meta = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> ext  = (int) ntohl (*netdata); ++netdata;
              self -> receiving -> len  = (int) ntohl (*netdata); ++netdata;

              /* enforce client ID onto message from (to avoid spoofing) */
              self -> receiving -> from = self -> client.id;

              yfree (self -> receiving -> data);

              if (self -> receiving -> len == 0)
                {
                  /* got the (bodyless) message */
                  self -> receiving -> data = NULL;
                  messageDespatch (&(self -> client), self -> receiving);
                  self -> receiving = NULL;
                }
              else
                {
                  self -> receiving -> data =
                         ymalloc (self -> receiving -> len + 1);
                  self -> receiving -> data[self -> receiving -> len] = '\0';
                }
            }
        }
      else if (self -> received >= MESSAGE_HEADER_SIZE)
        {
          ssize_t ret = read (self -> fd, self -> receiving -> data
                                  + self -> received - MESSAGE_HEADER_SIZE,
                              self -> receiving -> len - (self -> received
                                                    - MESSAGE_HEADER_SIZE));
          if (ret < 1)
            {
              clientClose (&(self -> client));
              return;
            }

          self -> received += ret;
          if (self -> received == self -> receiving -> len + MESSAGE_HEADER_SIZE)
            {
              /* got the message */
              messageDespatch (&(self -> client), self -> receiving);
              self -> receiving = NULL;
            }
        }
    }
  if (causeMask & CONTROL_WATCH_WRITE)
    {
      if (self -> sent < MESSAGE_HEADER_SIZE)
        {
          /* attempt to write the rest of the header */
          ssize_t ret;
          char data[MESSAGE_HEADER_SIZE];
          uint32_t *netdata = (uint32_t *)data;
          struct iovec buffers[2];
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> to);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> from);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> cid);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> oid);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> op);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> meta);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> ext);
          (*(netdata++)) = htonl ((uint32_t) self -> sending -> len);

          buffers[0].iov_base = data + self -> sent;
          buffers[0].iov_len = MESSAGE_HEADER_SIZE - self -> sent;
          buffers[1].iov_base = self -> sending -> data;
          buffers[1].iov_len = self -> sending -> len;

          ret = writev (self -> fd, buffers, self -> sending -> len ? 2 : 1);

          if (ret == 0)
            {
              perror ("whilst writing header");
              Y_FATAL ("Cannot write anything");
            }

          self -> sent += ret;
        }
      else 
        {
          ssize_t ret;
          ret = write (self -> fd, self -> sending -> data + (self -> sent - MESSAGE_HEADER_SIZE),
                       self -> sending -> len - (self -> sent - MESSAGE_HEADER_SIZE));
          if (ret == 0)
            {
              perror ("whilst writing body");
              Y_FATAL ("Cannot write anything");
            }

          self -> sent += ret;
        }
      if (self -> sent == self -> sending -> len + MESSAGE_HEADER_SIZE)
        {
          /* message sent */
          messageDestroy (self -> sending);
          if (listCount (self -> sendQueue) > 0)
            { 
              self -> sent = 0;
              self -> sending = listPopFront (self -> sendQueue);
            }
          else
            {
              self -> sending = NULL;
              controlChangeFileDescriptorMask (self -> fd,
                                               CONTROL_WATCH_READ );
            }
        }
    }
}

static void
unixSocketReady (int fd, int causeMask, void *data_v)
{
  /* struct UnixData *data = data_v; */
  struct UnixClient *newClient;

  newClient = ymalloc (sizeof (struct UnixClient));
  newClient -> fd = accept (fd, NULL, NULL);
  newClient -> sending = NULL;
  newClient -> receiving = NULL;
  newClient -> sent = 0;
  newClient -> received = 0;
  newClient -> sendQueue = listCreate ();

  newClient -> client.c = &unixClientClass;

  clientRegister (&(newClient -> client));

  controlRegisterFileDescriptor (newClient -> fd, CONTROL_WATCH_READ,
                                 newClient, unixClientReady);
  
};

int
initialise (struct Module *self, int argc, const char **argv)
{
  struct UnixData *data = ymalloc (sizeof (struct UnixData));
  struct sockaddr_un sockaddr;
  int i, r;

  self -> name = "IPC: UNIX Domain Sockets";
  self -> data = data;
  data -> path = NULL;

  for (i=0; i<argc; ++i)
    {
      if (strncmp (argv[i], "socket=", 7) == 0
          && strlen (argv[i]+7) > 0)
        data -> path = ystrdup (argv[i]+7);
    }

  if (data -> path == NULL)
    data -> path = ystrdup ("/tmp/.Y-0");
  sockaddr.sun_family = AF_UNIX;
  strncpy (sockaddr.sun_path, data->path, 100);

  data -> listeningFD = socket (PF_UNIX, SOCK_STREAM, 0);

  if (data -> listeningFD < 0)
    {
      /* clean up */
      /* write error message to Y log */
      return 0;
    }
 
  unlink (data -> path);
  r = bind (data->listeningFD, (struct sockaddr *)&sockaddr, sizeof (sockaddr));

  if (r < 0)
    {
      /* clean up */
      /* write error message to Y log */
      return 0;
    }

  r = chmod (data->path, 0777);

  if (r < 0)
    {
      /* clean up */
      /* write error message to Y log */
      return 0;
    }

  r = listen (data->listeningFD, 64);

  if (r < 0)
    {
      /* clean up */
      /* write error message to Y log */
      return 0;
    }
  
  controlRegisterFileDescriptor (data->listeningFD, CONTROL_WATCH_READ,
                                 data, unixSocketReady);

  return 0;
}

int
finalise (struct Module *self)
{
  struct UnixData *data = self -> data;
  unlink (data -> path);
  yfree (data -> path);
  yfree (data);
  return 0;
}

/* arch-tag: 2679bc05-8215-4840-a073-12c1c35c6a3e
 */
